import { writable } from 'svelte/store';
import JSZip from 'jszip';


function getStoredProjects() {
  if (typeof window !== 'undefined') {
    const storedProjects = localStorage.getItem('userProjects');
    return storedProjects ? JSON.parse(storedProjects) : [];
  }
  return [];
}

function toggleIsSavePending() {
  isSavePending.update(value => !value);
}
const initialProjects = getStoredProjects();

export const userProjects = writable(initialProjects);
export const isSavePending = writable(false);

if (typeof window !== 'undefined') {
  userProjects.subscribe(projects => {
    localStorage.setItem('userProjects', JSON.stringify(projects));
  });
}


/**
 * @param {string} name 
 */
export function addProject(name) {
  userProjects.update(projects => {
    const maxId = projects.reduce((/** @type {number} */ max, /** @type {{ id: number; }} */ project) => project.id > max ? project.id : max, 0);
    const newProject = { id: maxId + 1, name };
    toggleIsSavePending();
    return [...projects, newProject];
  });
}

/**
 * @param {string} name 
 */
export function removeProject(name) {
  toggleIsSavePending();
  userProjects.update(projects => {
    return projects.filter((/** @type {{ name: string; }} */ project) => project.name !== name);
  });
}

/**
 * @param {number} id 
 */
export function removeProjectById(id) {
  toggleIsSavePending();
  userProjects.update(projects => {
    return projects.filter((/** @type {{ id: number; }} */ project) => project.id !== id);
  });
}


/**
 * @typedef {Object} Project
 * @property {string} name - Der Name des Projekts
 * @property {number} id - Die Projektid
 * very basic structure for the Projects
 */

/**
 * Saves to Projects to a zip file
 * @returns {Promise<void>}
 */
export const saveProjectsAsZip = async () => {
  toggleIsSavePending();

  /** @type {Project[]} */
  let projectArray = [];
  userProjects.subscribe(value => {
    projectArray = value;
  })();

  const zipBlob = await createZipFile(projectArray);
  const url = URL.createObjectURL(zipBlob);
  const a = document.createElement('a');
  a.href = url;
  a.download = 'projects.zip';

  document.body.appendChild(a);
  a.click();

  document.body.removeChild(a);
  URL.revokeObjectURL(url);
};

/**
 * @param {number} id - Die Projekt-ID
 * @returns {string | undefined} - Der Name des Projekts oder undefined, wenn nicht gefunden
 */
export function getProjectNameById(id) {
  let projectName;
  userProjects.subscribe((projects) => {
    const project = projects.find((/** @type {Project} */ p) => p.id === id);
    projectName = project ? project.name : undefined;
  })();
  return projectName;
}


/**
* Loads the Project List from a zip file
*/
export const loadProjectsFromZip = (/** @type {File} */ file) => {
  loadProjectList(file);
  toggleIsSavePending();
};

const loadProjectList = async (/** @type {File} */ file) => {
  const zip = new JSZip();
  zip.loadAsync(file).then(async (zipContent) => {
    /** @type {Project[]} */
    const projectArray = [];
    /** @type {any[]} */
    const promises = [];
    zipContent.forEach(async (relativePath, file) => {
      const folderName = relativePath.split("/")[0];
      if (file.dir) return;

      const fileName = file.name.split("/")[1];
      if (fileName === "data.json") {
        const promise = file.async("text").then(content => {
          try {
            const projectData = JSON.parse(content);
            projectArray.push({ name: folderName, id: projectData.id });
          } catch (error) {
            console.error(`Error while trying to load ${folderName}:`, error);
          }
        });
        promises.push(promise);
      }
    });
    await Promise.all(promises);
    userProjects.set(projectArray);
    alert("Project list loaded successfully");
  }).catch(error => {
    alert("Error while loading the project list from a file");
    console.error(error);
  });
}

/**
 * helper function that parse the project zip file and loads it into the userProject list
 */
export const saveProjectsOnServer = async () => {
  /** @type {Project[]} */
  let projectArray = [];
  userProjects.subscribe(value => {
    projectArray = value;
  })();

  const zipBlob = await createZipFile(projectArray);
  const formData = new FormData();
  formData.append("file", new Blob([zipBlob]), "projects.zip");

  try {
    const response = await fetch('http://127.0.0.1:8000/projects/', {
      method: "POST",
      body: formData,
    });

    if (!response.ok) {
      throw new Error("Failed to upload ZIP file");
    }

    toggleIsSavePending();
    console.log("ZIP file uploaded successfully");
  } catch (error) {
    console.error("Error uploading file:", error);
  }
}

/**
 * Function to load the project list from the server
 */
export const loadProjectsFromServer = async () => {
  try {
    const response = await fetch('http://127.0.0.1:8000/projects/', {
      method: "GET",
    });
    if (!response.ok) {
      throw new Error("Failed to download the project list");
    }
    const blob = await response.blob();
    const file = new File([blob], "projects.zip", { type: blob.type });
    await loadProjectList(file);
    toggleIsSavePending();
  } catch (error) {
    throw error;
  }
}


/**
 * helper function:
 * generates a zip file with all the project files, every project gets its own directory
 */
const createZipFile = async (/** @type {Project[]} */ projects) => {
  const zip = new JSZip();

  projects.forEach((project) => {
    const folder = zip.folder(project.name);

    if (folder) {
      const fileContent = JSON.stringify({ id: project.id }, null, 2);
      folder.file("data.json", fileContent);
    } else {
      console.error(`Could not create folder for project: ${project.name}`);
    }
  });

  return await zip.generateAsync({ type: "blob" });
};